@@ -1,32 +1,118 @@
 use crate::data;
-use failure::Error;
+use anyhow::Error;
 use std::collections::HashMap;
+use std::io;
+use std::io::Write;
 
-use crate::data::{Aggregate, DisplayConfig, Record};
+use crate::data::{Aggregate, DisplayConfig, Record, Value, ValueDisplay};
 use crate::pipeline::OutputMode;
 use crate::render::{RenderConfig, TerminalConfig, TerminalSize};
-use itertools::Itertools;
+use itertools::{intersperse, Itertools};
 use strfmt::{strfmt_map, FmtError, Formatter};
 
-pub trait Printer<O> {
-    fn print(&mut self, row: &O, display_config: &DisplayConfig) -> String;
-    fn final_print(&mut self, row: &O, display_config: &DisplayConfig) -> String {
+pub trait AggregatePrinter {
+    fn print(&mut self, row: &data::Aggregate, display_config: &DisplayConfig) -> String;
+    fn final_print(&mut self, row: &data::Aggregate, display_config: &DisplayConfig) -> String {
         self.print(row, display_config)
     }
 }
 
-struct LogFmtPrinter {}
+pub trait RecordPrinter {
+    fn print(
+        &mut self,
+        out: &mut dyn Write,
+        row: &Record,
+        display_config: &DisplayConfig,
+    ) -> io::Result<()>;
+
+    fn print_str(&mut self, row: &Record, display_config: &DisplayConfig) -> String {
+        let mut out = vec![];
+        self.print(&mut out, row, display_config).unwrap();
+        String::from_utf8(out).unwrap()
+    }
+}
+
+struct RecordFromRow<T>(T);
+
+impl<T: RowPrinter> RecordPrinter for RecordFromRow<T> {
+    fn print(
+        &mut self,
+        out: &mut dyn Write,
+        row: &Record,
+        display_config: &DisplayConfig,
+    ) -> io::Result<()> {
+        self.0
+            .print_row(display_config, out, Some(&row.raw), &mut row.data.iter())
+    }
+}
+
+struct PrintAggregateAsRows<T>(T);
+
+pub(crate) trait RowPrinter {
+    fn print_row(
+        &mut self,
+        display_config: &DisplayConfig,
+        out: &mut dyn Write,
+        raw: Option<&str>,
+        cols: &mut dyn Iterator<Item = (&String, &Value)>,
+    ) -> io::Result<()>;
+}
+
+impl<T: RowPrinter> AggregatePrinter for PrintAggregateAsRows<T> {
+    fn print(&mut self, _row: &Aggregate, _display_config: &DisplayConfig) -> String {
+        "data will be output once the computation is complete...".to_string()
+    }
+
+    fn final_print(&mut self, row: &Aggregate, display_config: &DisplayConfig) -> String {
+        let mut out = vec![];
+        let columns = &row.columns;
+        for row in &row.data {
+            let mut kv_pairs = columns
+                .iter()
+                .map(|c| (c, row.get(c).unwrap_or(&Value::None)));
+            self.0
+                .print_row(display_config, &mut out, None, &mut kv_pairs)
+                .expect("writing to string");
+            writeln!(&mut out).unwrap();
+        }
+        String::from_utf8(out).expect("invalid UTF-8 produced")
+    }
+}
+
+struct LogFmtPrinter;
+
+impl RowPrinter for LogFmtPrinter {
+    fn print_row(
+        &mut self,
+        display_config: &DisplayConfig,
+        out: &mut dyn Write,
+        _raw: Option<&str>,
+        cols: &mut dyn Iterator<Item = (&String, &Value)>,
+    ) -> io::Result<()> {
+        let columns = intersperse(cols.sorted().map(Some), None);
+
+        for col in columns {
+            match col {
+                Some((col, data)) => {
+                    write!(out, "{}={}", col, ValueDisplay::new(data, display_config))
+                }
+                None => write!(out, " "),
+            }?;
+        }
+        Ok(())
+    }
+}
 
 pub fn raw_printer(
     mode: &OutputMode,
     render_config: RenderConfig,
     terminal_config: TerminalConfig,
-) -> Result<Box<dyn Printer<data::Record> + Send>, Error> {
+) -> Result<Box<dyn RecordPrinter + Send>, Error> {
     // suppress warnings until I use these
     let _x = terminal_config.color_enabled;
     let _y = terminal_config.is_tty;
     match mode {
-        OutputMode::Logfmt => Ok(Box::new(LogFmtPrinter {})),
+        OutputMode::Logfmt => Ok(Box::new(RecordFromRow(LogFmtPrinter))),
         OutputMode::Legacy => Ok(Box::new(LegacyPrinter::new(render_config, terminal_config))),
         OutputMode::Json => Ok(Box::new(JsonPrinter {})),
         OutputMode::Format(format_str) => Ok(Box::new(FormatPrinter::new(format_str.to_owned())?)),
@@ -37,21 +123,14 @@ pub fn agg_printer(
     mode: &OutputMode,
     render_config: RenderConfig,
     terminal_config: TerminalConfig,
-) -> Result<Box<dyn Printer<data::Aggregate> + Send>, Error> {
+) -> Result<Box<dyn AggregatePrinter + Send>, Error> {
     match mode {
-        //OutputMode::Logfmt => Ok(Box::new(LogFmtPrinter {})),
+        OutputMode::Logfmt => Ok(Box::new(PrintAggregateAsRows(LogFmtPrinter))),
+        OutputMode::Format(format_str) => Ok(Box::new(PrintAggregateAsRows(FormatPrinter::new(
+            format_str.to_owned(),
+        )?))),
         OutputMode::Json => Ok(Box::new(JsonPrinter {})),
-        _ => Ok(Box::new(LegacyPrinter::new(render_config, terminal_config))),
-        //OutputMode::Format(format_str) => Ok(Box::new(FormatPrinter::new(format_str.to_owned())?)),
-    }
-}
-
-impl Printer<data::Record> for LogFmtPrinter {
-    fn print(&mut self, row: &data::Record, display_config: &DisplayConfig) -> String {
-        let columns = row.data.iter().sorted();
-        let mut kv_pairs =
-            columns.map(|(col, data)| format!("{}={}", col, data.render(display_config)));
-        kv_pairs.join(" ")
+        OutputMode::Legacy => Ok(Box::new(LegacyPrinter::new(render_config, terminal_config))),
     }
 }
 
@@ -67,13 +146,18 @@ impl LegacyPrinter {
     }
 }
 
-impl Printer<data::Record> for LegacyPrinter {
-    fn print(&mut self, row: &Record, _display_config: &DisplayConfig) -> String {
-        self.pretty_printer.format_record_as_columns(row)
+impl RecordPrinter for LegacyPrinter {
+    fn print(
+        &mut self,
+        out: &mut dyn Write,
+        row: &Record,
+        _display_config: &DisplayConfig,
+    ) -> io::Result<()> {
+        write!(out, "{}", self.pretty_printer.format_record_as_columns(row))
     }
 }
 
-impl Printer<data::Aggregate> for LegacyPrinter {
+impl AggregatePrinter for LegacyPrinter {
     fn print(&mut self, row: &Aggregate, _display_config: &DisplayConfig) -> String {
         self.pretty_printer.format_aggregate(row)
     }
@@ -91,43 +175,85 @@ impl FormatPrinter {
     }
 }
 
-pub fn strformat_record(fmtstr: &str, vars: &Record) -> Result<String, FmtError> {
+pub fn strformat_record<'a>(
+    fmtstr: &str,
+    vars: impl Fn(&str) -> &'a Value,
+    display_config: &DisplayConfig,
+) -> Result<String, FmtError> {
     let formatter = |mut fmt: Formatter| {
-        let v = match vars.data.get(fmt.key) {
-            Some(v) => v,
-            None => &data::Value::None,
-        };
-        fmt.str(v.to_string().as_str())
+        let v = vars(fmt.key);
+        fmt.str(&format!("{}", ValueDisplay::new(v, display_config)))
     };
+
     strfmt_map(fmtstr, &formatter)
 }
 
-impl Printer<data::Record> for FormatPrinter {
-    fn print(&mut self, row: &Record, _display_config: &DisplayConfig) -> String {
-        match strformat_record(&self.format_str, row) {
-            Ok(s) => s,
-            Err(e) => format!("{}", e),
-        }
+impl RowPrinter for FormatPrinter {
+    fn print_row(
+        &mut self,
+        display_config: &DisplayConfig,
+        out: &mut dyn Write,
+        _raw: Option<&str>,
+        cols: &mut dyn Iterator<Item = (&String, &Value)>,
+    ) -> io::Result<()> {
+        let cols: Vec<_> = cols.collect();
+        write!(
+            out,
+            "{}",
+            strformat_record(
+                &self.format_str,
+                |query| cols
+                    .iter()
+                    .find(|(k, _)| query == *k)
+                    .map(|(_k, v)| v)
+                    .unwrap_or(&&Value::None),
+                display_config,
+            )
+            .unwrap_or_else(|e| format!("{}", e))
+        )
+    }
+}
+
+impl RecordPrinter for FormatPrinter {
+    fn print(
+        &mut self,
+        out: &mut dyn Write,
+        row: &Record,
+        _display_config: &DisplayConfig,
+    ) -> io::Result<()> {
+        write!(
+            out,
+            "{}",
+            match strformat_record(
+                &self.format_str,
+                |k| row.data.get(k).unwrap_or(&Value::None),
+                _display_config
+            ) {
+                Ok(s) => s,
+                Err(e) => format!("{}", e),
+            }
+        )
     }
 }
 
 struct JsonPrinter {}
 
-impl Printer<data::Record> for JsonPrinter {
-    fn print(&mut self, row: &Record, _display_config: &DisplayConfig) -> String {
-        serde_json::to_string(row).unwrap()
+impl RecordPrinter for JsonPrinter {
+    fn print(
+        &mut self,
+        out: &mut dyn Write,
+        row: &Record,
+        _display_config: &DisplayConfig,
+    ) -> io::Result<()> {
+        serde_json::to_writer(out, row).expect("failed to format");
+        Ok(())
     }
 }
 
-impl Printer<data::Aggregate> for JsonPrinter {
-    fn print(&mut self, _row: &Aggregate, _display_config: &DisplayConfig) -> String {
-        "JSON will be dumped once the pipeline is complete...\n".to_string()
-    }
-
-    fn final_print(&mut self, row: &Aggregate, _display_config: &DisplayConfig) -> String {
-        let mut o = serde_json::to_string(&row.data).unwrap();
-        o.push('\n');
-        o
+impl AggregatePrinter for JsonPrinter {
+    fn print(&mut self, row: &Aggregate, _display_config: &DisplayConfig) -> String {
+        let out = serde_json::to_string(row).expect("failed to format");
+        out + "\n"
     }
 }
 
@@ -358,9 +484,16 @@ impl PrettyPrinter {
 mod tests {
     use super::*;
     use crate::data::*;
+    use crate::operator::parse::ParseJson;
     use crate::operator::*;
     use maplit::hashmap;
 
+    impl LegacyPrinter {
+        fn print_record(&mut self, record: &Record, _display_config: &DisplayConfig) -> String {
+            self.pretty_printer.format_record_as_columns(record)
+        }
+    }
+
     #[test]
     fn print_raw() {
         let rec = Record::new("Hello, World!\n");
@@ -371,7 +504,7 @@ mod tests {
         };
         let display_config = DisplayConfig { floating_points: 2 };
         let mut pp = LegacyPrinter::new(render_config, TerminalConfig::load());
-        assert_eq!(pp.print(&rec, &display_config), "Hello, World!");
+        assert_eq!(pp.print_record(&rec, &display_config), "Hello, World!");
     }
 
     #[test]
@@ -387,14 +520,14 @@ mod tests {
         };
         let mut pp = LegacyPrinter::new(render_config, TerminalConfig::load());
         assert_eq!(
-            pp.print(&rec, &display_config),
+            pp.print_record(&rec, &display_config),
             "[k1=5]     [k2=5.50]    [k3=str]"
         );
         let rec = Record::new(r#"{"k1": 955, "k2": 5.5000001, "k3": "str3"}"#);
         let parser = ParseJson::new(None);
         let rec = parser.process(rec).unwrap().unwrap();
         assert_eq!(
-            pp.print(&rec, &display_config),
+            pp.print_record(&rec, &display_config),
             "[k1=955]   [k2=5.50]    [k3=str3]"
         );
         let rec = Record::new(
@@ -403,14 +536,14 @@ mod tests {
         let parser = ParseJson::new(None);
         let rec = parser.process(rec).unwrap().unwrap();
         assert_eq!(
-            pp.print(&rec, &display_config),
+            pp.print_record(&rec, &display_config),
             "[k1=here is a amuch longer stsring]    [k2=5.50]    [k3=str3]"
         );
         let rec = Record::new(r#"{"k1": 955, "k2": 5.5000001, "k3": "str3"}"#);
         let parser = ParseJson::new(None);
         let rec = parser.process(rec).unwrap().unwrap();
         assert_eq!(
-            pp.print(&rec, &display_config),
+            pp.print_record(&rec, &display_config),
             "[k1=955]                               [k2=5.50]    [k3=str3]"
         );
     }
@@ -422,12 +555,15 @@ mod tests {
         let display_config = DisplayConfig { floating_points: 2 };
         let rec = parser.process(rec).unwrap().unwrap();
         let mut pp = FormatPrinter::new("{k1:>3} k2={k2:<10.3} k3[{k3}]".to_string()).unwrap();
-        assert_eq!(pp.print(&rec, &display_config), "  5 k2=5.5        k3[str]");
+        assert_eq!(
+            pp.print_str(&rec, &display_config),
+            "  5 k2=5.5        k3[str]"
+        );
         let rec = Record::new(r#"{"k1": 955, "k2": 5.5000001, "k3": "str3"}"#);
         let parser = ParseJson::new(None);
         let rec = parser.process(rec).unwrap().unwrap();
         assert_eq!(
-            pp.print(&rec, &display_config),
+            pp.print_str(&rec, &display_config),
             "955 k2=5.5        k3[str3]"
         );
         let rec = Record::new(
@@ -436,7 +572,7 @@ mod tests {
         let parser = ParseJson::new(None);
         let rec = parser.process(rec).unwrap().unwrap();
         assert_eq!(
-            pp.print(&rec, &display_config),
+            pp.print_str(&rec, &display_config),
             "here is a amuch longer stsring k2=5.5        k3[str3]"
         );
     }
@@ -463,7 +599,10 @@ mod tests {
                 color_enabled: false,
             },
         );
-        assert_eq!(pp.print(&rec, &display_config), "[k1=5][k2=5.50][k3=str]");
+        assert_eq!(
+            pp.print_record(&rec, &display_config),
+            "[k1=5][k2=5.50][k3=str]"
+        );
     }
 
     #[test]
